@@ -18,6 +18,8 @@ import tempfile
 import yaml
 import traceback
 import cv2
+import pytesseract
+import requests
 
 from .engine_util import locate_on_screen, screenshot, locate
 from .engine_util import LARGE_FONT, NORM_FONT, SMALL_FONT, REGION_PICK_VIEW, OFFSET_PICK_VIEW, SIMILARITY_PICK_VIEW, NAME_PICK_VIEW
@@ -42,15 +44,8 @@ try:
 except:
     from src.aai_engine_package.screenshot_taker import ScreenShotTaker
 
-import logging
 import zmq
 
-logging.basicConfig(
-        format='%(asctime)s: %(levelname)s %(message)s',
-        level=logging.INFO,
-        datefmt='%H:%M:%S'
-    )
-
 REQUEST_TIMEOUT = 2500
 REQUEST_RETRIES = 3
 
@@ -58,6 +53,9 @@ ZMQ_PORT = 5555
 ZMQ_MSG_TYPE_PROGRESS = "progress"
 ZMQ_MSG_TYPE_DATA = "data"
 
+NORMAL_SCREENSHOT = 0
+OCR_SCREENSHOT = 1
+
 pyautogui.FAILSAFE = True
 
 CWD = os.getcwd()
@@ -108,13 +106,25 @@ def internal_display_used_as_main_on_osx() -> bool:
         return internal_display_as_main
 
 
-def get_zmq_client_socket():
-    """Establishes ZMQ connection with engine manager process and returns Socket"""
-    logging.info("Connecting to ZMQ client socket…")
-    context = zmq.Context()
-    socket = context.socket(zmq.REQ)
-    socket.connect(f"tcp://localhost:{ZMQ_PORT}")
-    return socket
+
+def init_dir(dirname):
+    """Initializes directory with given name"""
+    if not os.path.isdir(dirname):
+        original_umask = os.umask(0)
+        try:
+            os.makedirs(dirname, 0o777)
+        finally:
+            os.umask(original_umask)
+
+
+def log(filename, line, time=True):
+    """Log to file"""
+    with open(filename, "a+", encoding="utf-8") as file:
+        if time:
+            now = datetime.strftime(datetime.now(), '%H:%M:%S') + ": "
+        else:
+            now = ""
+        file.write(now + line + "\n")
 
 
 class TaskWrapper():
@@ -122,7 +132,7 @@ class TaskWrapper():
     Task wrapper class to control execution
     """
 
-    def __init__(self, task_id, name, cwd, script, scheduled_time, execution_type=-1, trigger=lambda _: True):
+    def __init__(self, task_id, name, cwd, script, scheduled_time, execution_type=-1, priority=15, trigger=lambda _: True):
         self.task_id = task_id
         self.name = name
         self.cwd = cwd
@@ -131,10 +141,7 @@ class TaskWrapper():
         self.scheduled_time = scheduled_time
         self.execution_type = execution_type
         self.trigger = trigger
-        self.log_dir = os.path.join(cwd, 'logs', self.name)
-        self.log_time = None
-
-        self.init_log_dir()
+        self.priority = priority
 
     def execute(self):
         """_summary_
@@ -150,20 +157,22 @@ class TaskWrapper():
         # info.dwFlags = subprocess.STARTUPINFO()
         # info.wShowWindow = SW_MINIMIZE
 
-        self.log_time = datetime.strftime(datetime.now(), '%Y%m%d-%H%M%S')
-        log_file_stdout = os.path.join(self.log_dir, self.log_time + "stdout.txt")
-        log_file_stderr = os.path.join(self.log_dir, self.log_time + "stdinfo.txt")
+        log_time = datetime.strftime(datetime.now(), '%Y%m%d-%H%M%S')
+        log_dir = os.path.join(self.cwd, "logs", self.name, log_time)
+        stdout_filename = os.path.join(log_dir, "stdout.txt")
+        zmq_log_filename = os.path.join(log_dir, "zmq_log.txt")
+        execution_log_filename = os.path.join(log_dir, "execution_log.txt")
+        init_dir(log_dir)
 
-        with open(log_file_stdout, "w+", encoding="utf-8") as stdout_log_file, open(log_file_stderr, "w+", encoding="utf-8") as stderr_log_file:
+        with open(stdout_filename, "w+", encoding="utf-8") as stdout_logfile:
             # check if python command exists on current machine and has version 3.x, otherwise run python3
             status, output = subprocess.getstatusoutput('python --version')
-            if status == 0 and output.startswith("Python 3"):
-                return subprocess.Popen(
-                    ['python', filepath], cwd=self.cwd, stdout=stdout_log_file,
-                    stderr=stderr_log_file, shell=False) #, startupinfo=info)
-
-            return subprocess.Popen(['python3', filepath], cwd=self.cwd, stdout=stdout_log_file, stderr=stderr_log_file, shell=False) #, startupinfo=info)
-
+            python_command = "python" if status == 0 and output.startswith("Python 3") else "python3"
+            process = subprocess.Popen([python_command, filepath, log_dir],  # pass log_dir as command line arg
+                                       cwd=self.cwd,
+                                       stdout=stdout_logfile,
+                                       shell=False)
+            return process, stdout_filename, zmq_log_filename, execution_log_filename
     def __str__(self) -> str:
         """str format for TaskWrapper"""
         if self.execution_type == 'SC':
@@ -172,17 +181,6 @@ class TaskWrapper():
             return f"ID:{self.task_id} = {self.name}||scheduled at: manually"
         return f"ID:{self.task_id} = {self.name}||scheduled at: continuously"
 
-    def init_log_dir(self):
-        """Init log directory
-        """
-        if not os.path.isdir(self.log_dir):
-            original_umask = os.umask(0)
-            try:
-                os.makedirs(self.log_dir, 0o777)
-            finally:
-                os.umask(original_umask)
-
-
 class Task():
     """
     Task class to control and keep track of all information and steps within a task.
@@ -200,18 +198,26 @@ class Task():
         self.cwd = cwd
         self.steps = []
         self.script = script
-        self.execution_log_file = None
-        self.log_time = datetime.strftime(datetime.now(), '%Y%m%d-%H%M%S')
-        self.execution_log_file = os.path.join(cwd, 'logs', self.name , self.log_time + "execution_log.txt")
-        self.standalone = standalone
         self._data = {}  # task-specific data
-        self.zmq_client_socket = get_zmq_client_socket()
+        log_time = datetime.strftime(datetime.now(), '%Y%m%d-%H%M%S')
+        self.log_dir = sys.argv[1] if not standalone else os.path.join(cwd, 'logs', name, log_time)
+        self.execution_logfile = os.path.join(self.log_dir, "execution_log.txt")
+        self.zmq_logfile = os.path.join(self.log_dir, "zmq_log.txt")
+        self.zmq_client_socket = self._get_zmq_client_socket()
         self.last_box_location = None
         self.error = False
+        self.standalone = standalone
+        # read execution and zmq logfile names from command line args if not run as standalone
 
-        self._init_log_dir()
         self._print_title()
 
+    def _get_zmq_client_socket(self):
+        """Establishes ZMQ connection with engine manager process and returns Socket"""
+        self._log_zmq("Connecting to ZMQ client socket.")
+        context = zmq.Context()
+        socket = context.socket(zmq.REQ)
+        socket.connect(f"tcp://localhost:{ZMQ_PORT}")
+        return socket
 
     def add_step(self, step):
         """Add a step to the current list of steps.
@@ -231,7 +237,8 @@ class Task():
 
         if self.script is None:
             self.error = False
-            self._update_progress(f"0/{len(self.steps)}")
+            if not self.standalone:
+                self._update_progress(f"0/{len(self.steps)}")
             for idx, step in enumerate(self.steps):
                 self._print_step(step,idx)
                 step.execute()
@@ -239,7 +246,7 @@ class Task():
                     self._update_progress(f"{idx + 1}/{len(self.steps)}")
             if not self.error:
                 print(f"{datetime.strftime(datetime.now(), '%H:%M:%S')}: TASK SUCCESSFULLY FINISHED!\n")
-                self._log("TASK SUCCESSFULLY FINISHED!\n")
+                self._log_execution("TASK SUCCESSFULLY FINISHED!\n")
             if not self.standalone:
                 # task finished, send task data as JSON string to engine manager process
                 # TODO maybe don't send task data immediately but allow control for when this happens outside of this function?
@@ -256,15 +263,15 @@ class Task():
         """Function that logs the most important things after a failure
         """
         self.error = True
-        screenshot_path = os.path.join(self.cwd, 'logs', self.name, self.log_time + ".png")
+        screenshot_path = os.path.join(self.log_dir, "last_screenshot.png")
         try:
             img = PngImageFile(os.path.join(self.cwd,".temp_screenshot.png"))
             img.save(screenshot_path)
         except Exception as ex:
             screenshot_path = "No screenshot could be found"
-        self._log(f"The last screenshot can be found at: {screenshot_path}", time=False)
-        self._log(f"The last box was: {self.last_box_location}", time=False)
-        self._log(f"TASK FAILED!")
+        self._log_execution(f"The last screenshot can be found at: {screenshot_path}", time=False)
+        self._log_execution(f"The last box was: {self.last_box_location}", time=False)
+        self._log_execution(f"TASK FAILED!")
 
     def _set_last_box_location(self, new_box_location):
         """Setter for the last box location used for logging purposes
@@ -281,9 +288,9 @@ class Task():
             step (Step): The step we want to log
             idx (int): Step number
         """
-        self._log("-"*(len(self.name) + FILLING_SPACE), time=False)
-        self._log("\tExecuting step ({idx}/{total}) - {name}".format(idx=(idx+1), total=len(self.steps), name=step.name), time=False)
-        self._log("-"*(len(self.name) + FILLING_SPACE), time=False)
+        self._log_execution("-"*(len(self.name) + FILLING_SPACE), time=False)
+        self._log_execution("\tExecuting step ({idx}/{total}) - {name}".format(idx=(idx+1), total=len(self.steps), name=step.name), time=False)
+        self._log_execution("-"*(len(self.name) + FILLING_SPACE), time=False)
 
     def _print_title(self):
         """Log the title of the task in the execution logs
@@ -292,29 +299,23 @@ class Task():
         s_char = "="
         str1 = s_char*int((FILLING_SPACE - len("Task Name: "))/2)
         str2 = s_char*title_length
-        self._log(str2, time=False)
-        self._log(str1 + "Task Name: " + self.name + str1, time=False)
-        self._log(str2, time=False)
+        self._log_execution(str2, time=False)
+        self._log_execution(str1 + "Task Name: " + self.name + str1, time=False)
+        self._log_execution(str2, time=False)
 
-    def _init_log_dir(self):
-        """Initializes the logging dir
-        """
-        if not os.path.isdir(os.path.join(self.cwd, 'logs', self.name)):
-            original_umask = os.umask(0)
-            try:
-                os.makedirs(os.path.join(self.cwd, 'logs', self.name), 0o777)
-            finally:
-                os.umask(original_umask)
+    def _log_execution(self, line, time=True):
+        """Log the execution process"""
+        log(self.execution_logfile, line, time)
 
-    def _log(self, line, time=True):
-        """Log the execution process
-        """
-        with open(self.execution_log_file, "a+", encoding="utf-8") as file:
-            if time:
-                now = datetime.strftime(datetime.now(), '%H:%M:%S') + ": "
-            else:
-                now = ""
-            file.write(now + line + "\n")
+    def _log_zmq(self, line, time=True):
+        """Log ZMQ communication"""
+        log(self.zmq_logfile, line, time)
+
+    def _send_zmq_msg(self, type, msg):
+        msg = {"type": type, "msg": msg}
+        self.zmq_client_socket.send_json(msg, ensure_ascii=False)
+        reply = self.zmq_client_socket.recv()
+        self._log_zmq(f"Server replied: {reply}")
 
     def _update_progress(self, progress):
         """Update progress on task
@@ -322,25 +323,13 @@ class Task():
         Args:
             progress (str): Description of the current progress
         """
-        request = json.dumps({"type": ZMQ_MSG_TYPE_PROGRESS, "msg": progress}, ensure_ascii=False).encode('utf8')
-        logging.info("Sending (%s)", request)
-        self.zmq_client_socket.send(request)
-        logging.info("Waiting for reply")
-        while True:
-            logging.info("Polling server")
-            if (self.zmq_client_socket.poll(REQUEST_TIMEOUT) & zmq.POLLIN) != 0:
-                reply = self.zmq_client_socket.recv()
-                logging.info("Server replied OK (%s)", reply)
-
-                break
+        self._log_zmq(f"Sending progress: {progress}")
+        self._send_zmq_msg(ZMQ_MSG_TYPE_PROGRESS, progress)
 
     def _send_task_data(self):
         '''Send task data to ZMQ server'''
-        logging.info("Sending task data")
-        msg = {"type": ZMQ_MSG_TYPE_DATA, "msg": self._data}
-        self.zmq_client_socket.send_json(msg, ensure_ascii=False)
-        reply = self.zmq_client_socket.recv()
-        logging.info(f"Server replied: {reply}")
+        self._log_zmq(f"Sending task data: {self._data}")
+        self._send_zmq_msg(ZMQ_MSG_TYPE_DATA, self._data)
 
     def set_data_item(self, key, val):
         """Set key-value pair in task data"""
@@ -359,30 +348,30 @@ class Task():
         """Calls engine.click
         """
         click(img_location, task=self)
-        self._log(f"-> click on {img_location} located at {self.last_box_location}")
+        self._log_execution(f"-> click on {img_location} located at {self.last_box_location}")
 
     def click_right(self, img_location):
         """Calls engine.click_right
         """
         click_right(img_location, task=self)
-        self._log(f"-> right click on {img_location} located at {self.last_box_location}")
+        self._log_execution(f"-> right click on {img_location} located at {self.last_box_location}")
 
     def double_click(self, img_location):
         """Calls engine.double_click
         """
         double_click(img_location, task=self)
-        self._log(f"-> double click on {img_location} located at {self.last_box_location}")
+        self._log_execution(f"-> double click on {img_location} located at {self.last_box_location}")
 
     def click_n(self, img_location, nr_clicks, button='left'):
         """Calls engine.click_n
         """
         click_n(img_location, nr_clicks, button=button, task=self)
-        self._log(f"-> click {button}, {nr_clicks} times on {img_location} located at {self.last_box_location}")
+        self._log_execution(f"-> click {button}, {nr_clicks} times on {img_location} located at {self.last_box_location}")
 
     def exists(self, img_location):
         """Calls engine.exists
         """
-        self._log("-> check if image exists")
+        self._log_execution("-> check if image exists")
         return exists(img_location, task=self)
 
     def get_box_location(self, img_location,haystack=None,iterations=5,steps_per_iteration=4,cache_confidence=True,debug=False):
@@ -393,31 +382,31 @@ class Task():
     def check_boxes_list(self, boxes,limit=5, treshold=10):
         """Calls engine.check_box_list
         """
-        self._log("-> Verify how close boxes are")
+        self._log_execution("-> Verify how close boxes are")
         return check_boxes_list(boxes, limit=limit, treshold=treshold, task=self)
 
     def wait(self,img_location, seconds):
         """Calls engine.wait
         """
-        self._log(f"-> Wait for {seconds}s, then try to find the image")
+        self._log_execution(f"-> Wait for {seconds}s, then try to find the image")
         wait(img_location, seconds, task=self)
 
     def sleep(self, seconds):
         """Calls engine.sleep
         """
-        self._log(f"-> sleep for {seconds}s")
+        self._log_execution(f"-> sleep for {seconds}s")
         sleep(seconds, task=self)
 
     def type_text(self, text):
         """Calls engine.type_text
         """
-        self._log(f"-> Write '{text}'")
+        self._log_execution(f"-> Write '{text}'")
         type_text(text, task=self)
 
     def key_combo(self, *keys):
         """Calls engine.key_combo
         """
-        self._log(f"-> click_right")
+        self._log_execution(f"-> click_right")
         key_combo(*keys, task=self)
 
     def remove_char(self, nr_characters=1):
@@ -428,25 +417,30 @@ class Task():
     def copy_to_clipboard(self):
         """Calls engine.copy_to_clipboard
         """
-        self._log(f"-> Copy to clipboard")
+        self._log_execution(f"-> Copy to clipboard")
         copy_to_clipboard(task=self)
 
     def get_clipboard(self):
         """Calls engine.get_clipboard
         """
-        self._log(f"-> Get clipboard contents")
+        self._log_execution(f"-> Get clipboard contents")
         return get_clipboard(task=self)
 
     def read_excel(self, path):
         """Calls engine.read_excel
         """
-        self._log(f"-> Read excel file")
+        self._log_execution(f"-> Read excel file")
         result = read_excel(path, task=self)
-        self._log(f"\tDATA from {path}: ", time=False)
+        self._log_execution(f"\tDATA from {path}: ", time=False)
         for point in result:
-            self._log("\t\t" + str(point), time=False)
+            self._log_execution("\t\t" + str(point), time=False)
         return result
 
+    def ocr_image(self, img_location, haystack_path = None, on_needle=False):
+        """Calls engine.ocr_image
+        """
+        ocr_image(img_location, haystack_path=haystack_path, on_needle=on_needle)
+
 
 class Step():
     """
@@ -465,11 +459,11 @@ class Step():
         try:
             self.func(self.task, *self.args)
         except Exception as ex:
-            self.task._log("ERROR: ")
-            self.task._log(traceback.format_exc(), time=False)
+            self.task._log_execution("ERROR: ")
+            self.task._log_execution(traceback.format_exc(), time=False)
             self.task._task_failed()
 
-    def set_task(self, task):
+    def set_task(self, task: Task):
         """Set the step's corresponding Task"""
         self.task = task
 
@@ -702,6 +696,84 @@ def check_boxes_list(boxes,limit=5, treshold=10, task=None):
 
     return None
 
+def ocr_image_server(url, img_location, haystack_path = None, on_needle=False):
+    """Perform OCR on an image on the server
+
+    Args:
+        url (str): url for the server request
+        img_location (str): Location of the needle image
+        haystack_path (str, optional): Path to a self-specified haystack. If None a new screenshot will be taken. Defaults to None.
+        on_needle (bool, optional): Perform OCR on the needle image (True) or the OCR box, given in the metadata of the image (False). Defaults to False.
+
+    Returns:
+        str: Tesseract output of the text in the image
+    """
+    try:
+        needle = PngImageFile(img_location)
+        metadata = needle.text
+        if haystack_path:
+            haystack = PngImageFile(haystack_path)
+        else:
+            haystack = screenshot(CWD + "/.temp_screenshot.png")
+        anchor_box = get_box_location(img_location, haystack=haystack)
+
+        ocr_box = [2*int(i) for i in metadata['ocr_box_relative'][1:-1].split(',')]
+
+        ocr_img = haystack.crop((anchor_box.left + ocr_box[0],anchor_box.top + ocr_box[1],anchor_box.left + ocr_box[0] + ocr_box[2],anchor_box.top + ocr_box[1] + ocr_box[3]))
+
+        ocr_img.save(CWD + "/.temp_ocr.png")
+
+        # Ask server to do OCR on the image
+        response = requests.post(url, files={'image':open(CWD + "/.temp_ocr.png", "rb")}, data=needle.text)
+        return response.text
+    except Exception as ex:
+        print(f"{datetime.strftime(datetime.now(), '%H:%M:%S')}: Could not load the needle and haystack image! \n{ex}")
+    return None
+
+def ocr_image(img_location, haystack_path = None, on_needle=False):
+    """Perform OCR on a given image.
+
+    Args:
+        img_location (str): Location of the needle image
+        haystack_path (str, optional): Path to a self-specified haystack. If None a new screenshot will be taken. Defaults to None.
+        on_needle (bool, optional): Perform OCR on the needle image (True) or the OCR box, given in the metadata of the image (False). Defaults to False.
+
+    Returns:
+        str: Tesseract output of the text in the image
+    """
+    print(f"{datetime.strftime(datetime.now(), '%H:%M:%S')}: Perform OCR on image")
+    if on_needle:
+        # Tesseract puts unnecessary characters at the end of the text -> solution = [:-2]
+        try:
+            return pytesseract.image_to_string(Image.open(img_location))[:-2]
+        except Exception as ex:
+            print(f"{datetime.strftime(datetime.now(), '%H:%M:%S')}: OCR failed! \n{ex}")
+    else:
+        # Get the anchor in the haystack
+        try:
+            if not haystack_path:
+                haystack = screenshot(CWD + "/.temp_screenshot.png")
+            else:
+                haystack = PngImageFile(haystack_path)
+            anchor_box = get_box_location(img_location, haystack=haystack)
+        except Exception as ex:
+            print(f"{datetime.strftime(datetime.now(), '%H:%M:%S')}: Anchor could not be found! \n{ex}")
+
+        # Crop the haystack to the OCR box
+        needle = PngImageFile(img_location)
+        img_text = needle.text
+        # Weird behavior -> Box must be twice as large
+        ocr_box = [2*int(i) for i in img_text['ocr_box_relative'][1:-1].split(',')]
+        # print(f"Anchor: {anchor_box}")
+        # print(f"OCR: {ocr_box}")
+        try:
+            ocr_img = haystack.crop((anchor_box.left + ocr_box[0],anchor_box.top + ocr_box[1],anchor_box.left + ocr_box[0] + ocr_box[2],anchor_box.top + ocr_box[1] + ocr_box[3]))
+            # ocr_img.show()
+            return pytesseract.image_to_string(ocr_img)[:-2]
+        except Exception as ex:
+            print(f"{datetime.strftime(datetime.now(), '%H:%M:%S')}: OCR failed! \n{ex}")
+    return None
+
 
 def wait(img_location, seconds, task=None):
     """ Wait a given amount of seconds for a given image, checking its existence.
@@ -827,7 +899,7 @@ def cli():
         save_location = sys.argv[1]
         print(os.getcwd(), save_location)
     print("Called aai-engine-capture")
-    main(r"H:\AdAstraIndustries\aai_engine\img")
+    main(save_location)
 
 def test_edit(save_location, param2):
     """Edit stub
@@ -841,25 +913,18 @@ def edit(save_location, haystack):
     main(save_location, haystack, editing=True)
 
 
-def main(save_location, needle=r"C:\Users\Toto\Documents\AdAstraIndustries\aai_engine\img\cv.png", editing=False):
+def main(save_location, update_function_after_edit=None, needle=r"C:\Users\Toto\Documents\AdAstraIndustries\aai_engine\img\cv.png", editing=False, screenshot_type = NORMAL_SCREENSHOT):
     """main
     """
     temp_screenshot = tempfile.NamedTemporaryFile(suffix='.png', delete=False)
     temp_screenshot_path = temp_screenshot.name
-    if os.name == "posix":
+    if sys.platform == "darwin":
         temp_screenshot_path = temp_screenshot_path.split(".")[0] + "_000.png"
     print("TEMP PATH: ", temp_screenshot_path)
     take_screenshot(temp_screenshot_path)
-    # take_screenshot()
-    # TODO: make title bar black
-    # aai_window = AAIWindow()
-    # root = aai_window.root
 
     root = tkinter.Tk()
     style = ttk.Style(root)
-    # iconfile = pkg_resources.read_binary(style_dir, 'icon.ico')
-    # iconfile = pkgutil.get_data(__name__, "style/icon.ico")
-    # root.wm_iconbitmap(icon_path)
     root.wm_colormapwindows()
 
     root.tk.call('source', theme_path)
@@ -867,11 +932,11 @@ def main(save_location, needle=r"C:\Users\Toto\Documents\AdAstraIndustries\aai_e
 
     if editing:
         print(" - EDIT MODE - ")
-        app = ScreenShotTaker(root, save_location, needle=needle, haystack=temp_screenshot_path,
+        app = ScreenShotTaker(root, save_location, update_function_after_edit=update_function_after_edit, needle=needle, haystack=temp_screenshot_path,
                               editing=True)  # When editing from extension
     else:
         print(" - CREATE MODE - ")
-        app = ScreenShotTaker(root, save_location, haystack=temp_screenshot_path, editing=False)
+        app = ScreenShotTaker(root, save_location, update_function_after_edit=update_function_after_edit, haystack=temp_screenshot_path, editing=False, screenshot_type=screenshot_type)
 
     root.mainloop()
     temp_screenshot.close()
